/*----------------------------------------------------------------------
 * This source distribution and its contents are (c) 2011 Gregory Perry, 
 * and are released under the GNU Public License Version 3.  Portions of
 * this distribution were borrowed from the msieve project.
 *
 * For more information visit the main CloudCrack page at:
 * http://www.GoVirtual.tv/CloudCrack/
 *
 * Become of fan and engage in discussions at the CloudCrack Fan Page:
 * https://www.facebook.com/pages/CloudCrack/156984984362011/
 *
 * Or, send an email to Gregory.Perry@GoVirtual.tv
 *
 * $Id$
 * --------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

#define MAX_MP_WORDS 8

typedef signed int int32;
typedef unsigned int uint32;
typedef signed long long int64;
typedef unsigned long long uint64;

/* note that this struct must be aligned by 8 to avoid
   a compiler error */

typedef struct __align__(8) {
	uint32 mont_w;
	uint32 target[MAX_MP_WORDS];
} gpu_const_data_t;

__constant__ gpu_const_data_t const_data;

/*---------------------------------------------------------------------*/
__device__ int32
mont_cmp(uint32 *a, uint32 stride_a,
	 uint32 *b, uint32 stride_b)
{
	uint32 i, j;

	i = stride_a * (MAX_MP_WORDS - 1);
	j = stride_b * (MAX_MP_WORDS - 1);
	while ((int32)j >= 0) {
		uint32 ai = a[i];
		uint32 bj = b[j];

		if (ai > bj)
			return 1;
		else if (ai < bj)
			return -1;

		i -= stride_a;
		j -= stride_b;
	}

	return 0;
}

/*---------------------------------------------------------------------*/
__device__ uint32
mont_add(uint32 *a, uint32 stride_a,
		uint32 *b, uint32 stride_b,
		uint32 *c, uint32 stride_c)
{
	uint32 i, j, k, m;
	int64 carry = 0;

	for (i = j = k = m = 0; m < MAX_MP_WORDS; m++) {

		uint64 ai = a[i];
		uint64 bj = b[j];

		carry += ai + bj;
		c[k] = (uint32)carry;
		carry = carry >> 32;

		i += stride_a;
		j += stride_b;
		k += stride_c;
	}

	return carry;
}

/*---------------------------------------------------------------------*/
__device__ void
mont_sub(uint32 *a, uint32 stride_a,
		uint32 *b, uint32 stride_b,
		uint32 *c, uint32 stride_c)
{
	uint32 i, j, k, m;
	int64 carry = 0;

	for (i = j = k = m = 0; m < MAX_MP_WORDS; m++) {

		uint64 ai = a[i];
		uint64 bj = b[j];

		carry += ai - bj;
		c[k] = (uint32)carry;
		carry = carry >> 32;

		i += stride_a;
		j += stride_b;
		k += stride_c;
	}
}

/*---------------------------------------------------------------------*/
__device__ void 
montmul(uint32 *a, uint32 *b, uint32 *scratch, uint32 stride)
{
	uint32 i, j, k, m;
	uint32 *n = const_data.target;

	for (i = k = 0; i <= MAX_MP_WORDS; i++, k+=stride)
		scratch[k] = 0;

	for (i = k = 0; i < MAX_MP_WORDS; i++, k+=stride) {

		uint32 q;
		uint32 carry;
		uint64 accum;

		q = a[k];
		accum = (uint64)q * b[0] + scratch[0];
		scratch[0] = (uint32)accum;
		carry = (uint32)(accum >> 32);

		for (j = 1, m = stride; j < MAX_MP_WORDS; j++, m+=stride) {
			accum = (uint64)q * b[m] + carry + scratch[m];
			scratch[m] = (uint32)accum;
			carry = (uint32)(accum >> 32);
		}

		scratch[m] += carry;
		q = scratch[0] * const_data.mont_w;
		accum = (uint64)q * n[0] + scratch[0];
		carry = (uint32)(accum >> 32);

		for (j = 1, m = stride; j < MAX_MP_WORDS; j++, m+=stride) {
			accum = (uint64)q * n[j] + carry + scratch[m];
			scratch[m - stride] = (uint32)accum;
			carry = (uint32)(accum >> 32);
		}

		accum = (uint64)scratch[m] + carry;
		scratch[m - stride] = (uint32)accum;
		scratch[m] = (uint32)(accum >> 32);
	}

	if (scratch[m] || mont_cmp(scratch, stride, n, 1) >= 0)
		mont_sub(scratch, stride, n, 1, scratch, stride);
}

/*---------------------------------------------------------------------*/
__device__ uint32
mont_is_zero(uint32 *a, uint32 stride_a)
{
	uint32 i, j;

	for (i = j = 0; i < MAX_MP_WORDS; i++, j += stride_a) {
		if (a[j] != 0)
			return 0;
	}

	return 1;
}

/*---------------------------------------------------------------------*/
__device__ void
gcd(uint32 *a, uint32 *b, uint32 stride)
{
	uint32 i, j;
	uint32 *u = a;
	uint32 *v = b;

	if (mont_is_zero(v, stride)) {
		return;
	}
	else if (v[0] & 1) {
		u = b;
		v = a;
		mont_sub(v, stride, u, stride, v, stride);
	}

	do {
		do {
			uint32 shift, comp_shift;
			uint32 lo, hi;

#if 0
			shift = __ffs(v[0]) - 1; /* why doesn't this work? */
			if (shift > 31)
				shift = 31;
#else
			j = v[0];
			shift = 0;
			if (j == 0) {
				shift = 31;
			}
			else {
				while ((j & 1) == 0) {
					shift++;
					j >>= 1;
				}
			}

#endif
			comp_shift = 32 - shift;
			lo = v[0];
			for (i = j = 0; i < MAX_MP_WORDS-1; i++, j+=stride) {
				hi = v[j+stride];
				v[j] = lo >> shift | hi << comp_shift;
				lo = hi;
			}
			v[j] = lo >> shift;

		} while ((v[0] & 1) == 0 && !mont_is_zero(v, stride));

		if (mont_cmp(u, stride, v, stride) > 0) {
			uint32 *tmp = u;
			u = v;
			v = tmp;
		}

		mont_sub(v, stride, u, stride, v, stride);

	} while (!mont_is_zero(v, stride));

	if (mont_is_zero(b, stride)) {
		for (i = j = 0; i < MAX_MP_WORDS; i++, j+=stride)
			b[j] = a[j];
	}
}

/*---------------------------------------------------------------------*/
__global__ void 
brent_kernel(uint32 base_iter, uint32 base_limit,
		uint32 num_iter, uint32 num_functions,
		uint32 *work_array)
{
	uint32 j, k, m;
	uint32 n = num_functions;

	uint32 i = blockIdx.x * blockDim.x + threadIdx.x;
	uint32 *tmp = work_array + i;
	uint32 *accum = work_array + MAX_MP_WORDS * n + i;
	uint32 *inc = work_array + 2 * MAX_MP_WORDS * n + i;
	uint32 *x = work_array + 3 * MAX_MP_WORDS * n + i;
	uint32 *y = work_array + 4 * MAX_MP_WORDS * n + i;
	uint32 *mont = work_array + 5 * MAX_MP_WORDS * n + i;

	uint32 *target = const_data.target;
	uint32 curr_iter = base_iter;
	uint32 curr_limit = base_limit;
	uint32 parity = 0;

	for (j = 0; j < num_iter; j++, curr_iter++) {

		uint32 *sub0;
		uint32 *sub1;
		uint32 carry;

		montmul(x, x, mont, n);

		carry = mont_add(mont, n, inc, n, x, n);

		if (carry || mont_cmp(x, n, target, 1) >= 0)
			mont_sub(x, n, target, 1, x, n);

		sub0 = y;
		sub1 = x;
		if (mont_cmp(y, n, x, n) < 0) {
			sub0 = x;
			sub1 = y;
			parity ^= 1;
		}
		mont_sub(sub0, n, sub1, n, tmp, n);

		montmul(tmp, accum, mont, n);

		for (k = m = 0; k < MAX_MP_WORDS; k++, m+=n)
			accum[m] = mont[m];

		if (curr_iter == curr_limit) {
			for (k = m = 0; k < MAX_MP_WORDS; k++, m+=n)
				y[m] = x[m];
			curr_limit *= 2;
		}
	}
	
	if (parity)
		mont_sub(target, 1, accum, n, accum, n);
}

/*---------------------------------------------------------------------*/
__global__ void 
do_gcd(uint32 num_functions, uint32 *work_array)
{
	uint32 j;
	uint32 n = num_functions;
	uint32 i = blockIdx.x * blockDim.x + threadIdx.x;
	uint32 *tmp = work_array + i;
	uint32 *mont = work_array + 5 * MAX_MP_WORDS * n + i;
	uint32 *target = const_data.target;

	n = num_functions;
	for (i = j = 0; i < MAX_MP_WORDS; i++, j+=n)
		mont[j] = target[i];

	gcd(mont, tmp, n);
}


#ifdef __cplusplus
}
#endif
