/*----------------------------------------------------------------------
 * This source distribution and its contents are (c) 2011 Gregory Perry, 
 * and are released under the GNU Public License Version 3.  Portions of
 * this distribution were borrowed from the msieve project.
 *
 * For more information visit the main CloudCrack page at:
 * http://www.GoVirtual.tv/CloudCrack/
 *
 * Become of fan and engage in discussions at the CloudCrack Fan Page:
 * https://www.facebook.com/pages/CloudCrack/156984984362011/
 *
 * Or, send an email to Gregory.Perry@GoVirtual.tv
 *
 * $Id$
 * --------------------------------------------------------------------*/

#include "cudarho_priv.h"

extern int32 evaluate_expression(char *expr, mp_t *res); 

/*---------------------------------------------------------------------*/
#ifdef HAVE_CUDA
static void
gpu_rho_init(rho_t *r, uint32 which_gpu)
{
	gpu_config_t gpu_config;
	uint32 threads_per_block;
	mp_t gpu_target_data;
	void *gpu_ptr;
	uint32 i;

	/* find a GPU */

	gpu_init(&gpu_config);
	if (gpu_config.num_gpu == 0) {
		printf("error: no CUDA-enabled GPUs found\n");
		exit(-1);
	}
	if (which_gpu >= (uint32)gpu_config.num_gpu) {
		printf("error: GPU %u does not exist "
			"or is not CUDA-enabled\n", which_gpu);
		exit(-1);
	}
	printf("using GPU %u (%s)\n", which_gpu,
			gpu_config.info[which_gpu].name);

	r->gpu_info = gpu_config.info[which_gpu];

	/* set up computing context */

	CUDA_TRY(cuCtxCreate(&r->gpu_context, 
			CU_CTX_BLOCKING_SYNC,
			r->gpu_info.device_handle))

	/* load and compile the kernel */

	CUDA_TRY(cuModuleLoad(&r->gpu_module, "brent.ptx"))

	CUDA_TRY(cuModuleGetFunction(&r->gpu_kernel, 
				r->gpu_module, "brent_kernel"))

	CUDA_TRY(cuModuleGetFunction(&r->gpu_kernel2, 
				r->gpu_module, "do_gcd"))

	/* point to and fill the kernel's array of constant data */

	CUDA_TRY(cuModuleGetGlobal(&r->gpu_const_data, 
				NULL, r->gpu_module, "const_data"))

	mp_copy(&r->target, &gpu_target_data);
	gpu_target_data.nwords = r->mont_w;

	CUDA_TRY(cuMemcpyHtoD(r->gpu_const_data, 
			&gpu_target_data, sizeof(mp_t)))

	/* figure out how large a chunk of work can be */

	CUDA_TRY(cuFuncGetAttribute((int *)&threads_per_block, 
			CU_FUNC_ATTRIBUTE_MAX_THREADS_PER_BLOCK,
			r->gpu_kernel))

	/* override the number of functions with a value that
	   is convenient to the card */

	r->num_functions = threads_per_block * ((r->num_functions + 
				(threads_per_block - 1)) / threads_per_block);

	CUDA_TRY(cuFuncSetBlockShape(r->gpu_kernel, 
				threads_per_block, 1, 1))

	CUDA_TRY(cuFuncSetBlockShape(r->gpu_kernel2, 
				threads_per_block, 1, 1))

	printf("GPU will run %u PRNGs\n", r->num_functions);

	r->num_blocks = r->num_functions / threads_per_block;

	/* allocate GPU memory for the work array (later
	   code needs to fill it with the iteration data) */

	CUDA_TRY(cuMemAlloc(&r->gpu_iter_data, (6 * MAX_MP_WORDS + 1) *
				r->num_functions * sizeof(uint32)))

	/* set all the constant function arguments the kernel
	   will use */

	i = 3 * sizeof(uint32);
	CUDA_TRY(cuParamSeti(r->gpu_kernel, i, r->num_functions))
	i += sizeof(uint32);

	gpu_ptr = (void *)(size_t)r->gpu_iter_data;
	CUDA_ALIGN_PARAM(i, __alignof(gpu_ptr));
	CUDA_TRY(cuParamSetv(r->gpu_kernel, (int)i, 
			&gpu_ptr, sizeof(gpu_ptr)))
	i += sizeof(gpu_ptr);

	CUDA_TRY(cuParamSetSize(r->gpu_kernel, i))


	i = 0;
	CUDA_TRY(cuParamSeti(r->gpu_kernel2, i, r->num_functions))
	i += sizeof(uint32);

	gpu_ptr = (void *)(size_t)r->gpu_iter_data;
	CUDA_ALIGN_PARAM(i, __alignof(gpu_ptr));
	CUDA_TRY(cuParamSetv(r->gpu_kernel2, (int)i, 
			&gpu_ptr, sizeof(gpu_ptr)))
	i += sizeof(gpu_ptr);

	CUDA_TRY(cuParamSetSize(r->gpu_kernel2, i))
}

#endif

/*---------------------------------------------------------------------*/
static uint32 
get_mont_w(uint32 n) 
{
	uint32 res = 2 + n;
	res = res * (2 + n * res);
	res = res * (2 + n * res);
	res = res * (2 + n * res);
	return res * (2 + n * res);
}

/*---------------------------------------------------------------------*/
static mp_t four = {1, {4}};

static void
get_mont_r(mp_t *n, mp_t *r)
{
	mp_t num, rem;

	mp_clear(&num);
	num.nwords = MAX_MP_WORDS;
	num.val[MAX_MP_WORDS-1] = 1 << 31;

	mp_mod(&num, n, &rem);
	mp_modmul(&rem, &rem, n, &rem);
	mp_modmul(&rem, &four, n, r);
}

/*---------------------------------------------------------------------*/
void rho_init(rho_t *r, char *input, 
		rng_callback_t prng_func, void *prng_data, 
		uint32 num_functions, uint32 which_gpu)
{
	uint32 i, j, k;
	int32 status;
	char buf[1000];
	char *target_string;

	memset(r, 0, sizeof(rho_t));

	status = evaluate_expression(input, &r->target);
	if (status < 0 || mp_is_zero(&r->target)) {
		printf("error %d converting '%s'\n", status, input);
		return;
	}
	target_string = mp_sprintf(&r->target, 10, buf);
	printf("factoring %s (%d digits)\n", 
			target_string, strlen(target_string));

	r->num_functions = num_functions;
	r->curr_iter = 2;
	r->curr_limit = 2;
	r->mont_w = get_mont_w(r->target.val[0]);
	get_mont_r(&r->target, &r->mont_r);

#ifdef HAVE_CUDA
	gpu_rho_init(r, which_gpu);
	num_functions = r->num_functions;
#endif
	r->curr_values = (mp_t *)xmalloc(num_functions * sizeof(mp_t));

	r->main_array = (uint32 *)xcalloc((6 * MAX_MP_WORDS + 1) * 
					num_functions, sizeof(uint32));

	r->gcd_array = r->main_array;
	r->accum_array = r->gcd_array + MAX_MP_WORDS * num_functions;
	r->inc_array = r->accum_array + MAX_MP_WORDS * num_functions;
	r->x_array = r->inc_array + MAX_MP_WORDS * num_functions;
	r->y_array = r->x_array + MAX_MP_WORDS * num_functions;
	r->mont_temp = r->y_array + MAX_MP_WORDS * num_functions;

	for (i = 0; i < num_functions; i++) {

		mp_t rand_input;

		for (j = k = 0; j < MAX_MP_WORDS; j++, k+=num_functions)
			r->accum_array[k+i] = r->mont_r.val[j];

		prng_func(i, prng_data, &rand_input);

		for (j = k = 0; j < MAX_MP_WORDS; j++, k+=num_functions)
			r->inc_array[k+i] = rand_input.val[j];

		montmul(r->inc_array + i, r->accum_array + i, 
			r->mont_temp + i, num_functions, r->target.val,
			r->mont_w);

		for (j = k = 0; j < MAX_MP_WORDS; j++, k+=num_functions)
			r->inc_array[k+i] = r->mont_temp[k+i];


		prng_func(i, prng_data, &rand_input);

		for (j = k = 0; j < MAX_MP_WORDS; j++, k+=num_functions)
			r->x_array[k+i] = rand_input.val[j];

		montmul(r->x_array + i, r->accum_array + i, 
			r->mont_temp + i, num_functions, r->target.val,
			r->mont_w);

		for (j = k = 0; j < MAX_MP_WORDS; j++, k+=num_functions)
			r->x_array[k+i] = r->y_array[k+i] = r->mont_temp[k+i];


		r->gcd_array[i] = 1;

		montmul(r->gcd_array + i, r->accum_array + i, 
			r->mont_temp + i, num_functions, r->target.val,
			r->mont_w);

		for (j = k = 0; j < MAX_MP_WORDS; j++, k+=num_functions)
			r->accum_array[k+i] = r->mont_temp[k+i];
	}

#ifdef HAVE_CUDA
	CUDA_TRY(cuMemcpyHtoD(r->gpu_iter_data, 
			r->main_array, (6 * MAX_MP_WORDS + 1) *
			num_functions * sizeof(uint32)))
#endif

}

/*---------------------------------------------------------------------*/
void rho_free(rho_t *r)
{
	free(r->main_array);
#ifdef HAVE_CUDA
	CUDA_TRY(cuMemFree(r->gpu_iter_data))
	CUDA_TRY(cuCtxDestroy(r->gpu_context)) 
#endif
	memset(r, 0, sizeof(rho_t));
}
