/*----------------------------------------------------------------------
 * This source distribution and its contents are (c) 2011 Gregory Perry, 
 * and are released under the GNU Public License Version 3.  Portions of
 * this distribution were borrowed from the msieve project.
 *
 * For more information visit the main CloudCrack page at:
 * http://www.GoVirtual.tv/CloudCrack/
 *
 * Become of fan and engage in discussions at the CloudCrack Fan Page:
 * https://www.facebook.com/pages/CloudCrack/156984984362011/
 *
 * Or, send an email to Gregory.Perry@GoVirtual.tv
 *
 * $Id$
 * --------------------------------------------------------------------*/

#include "cudarho_priv.h"

/*---------------------------------------------------------------------*/
#if !defined(HAVE_CUDA)
static void 
rho_run_cpu(rho_t *r, uint32 num_iterations, 
		uint32 base_iter, uint32 base_limit)
{
	uint32 i, j, k, m;
	uint32 n = r->num_functions;
	uint32 curr_iter;
	uint32 curr_limit;

	for (i = 0; i < n; i++) {

		uint32 parity = 0;
		uint32 *x = r->x_array + i;
		uint32 *y = r->y_array + i;
		uint32 *inc = r->inc_array + i;
		uint32 *mont = r->mont_temp + i;
		uint32 *accum = r->accum_array + i;
		uint32 *tmp = r->gcd_array + i;
		uint32 *target = r->target.val;
		uint32 w = r->mont_w;
		curr_iter = base_iter;
		curr_limit = base_limit;

		for (j = 0; j < num_iterations; j++, curr_iter++) {

			uint32 *sub0;
			uint32 *sub1;
			uint32 carry;

			montmul(x, x, mont, n, target, w);

			carry = mont_add(mont, n, inc, n, x, n);

			if (carry || mont_cmp(x, n, target, 1) >= 0)
				mont_sub(x, n, target, 1, x, n);

			sub0 = y;
			sub1 = x;
			if (mont_cmp(y, n, x, n) < 0) {
				sub0 = x;
				sub1 = y;
				parity ^= 1;
			}
			mont_sub(sub0, n, sub1, n, tmp, n);

			montmul(tmp, accum, mont, n, target, w);

			for (k = m = 0; k < MAX_MP_WORDS; k++, m+=n)
				accum[m] = mont[m];

			if (curr_iter == curr_limit) {
				for (k = m = 0; k < MAX_MP_WORDS; k++, m+=n)
					y[m] = x[m];
				curr_limit *= 2;
			}
		}
		
		if (parity)
			mont_sub(target, 1, accum, n, accum, n);
	}
}
#endif

/*---------------------------------------------------------------------*/
void 
rho_run(rho_t *r, uint32 num_iterations)
{
	uint32 base_iter = r->curr_iter;
	uint32 base_limit = r->curr_limit;

#ifdef HAVE_CUDA
	CUDA_TRY(cuParamSeti(r->gpu_kernel, 0, base_iter))
	CUDA_TRY(cuParamSeti(r->gpu_kernel, 4, base_limit))
	CUDA_TRY(cuParamSeti(r->gpu_kernel, 8, num_iterations))
	CUDA_TRY(cuLaunchGrid(r->gpu_kernel, r->num_blocks, 1))
#else
	rho_run_cpu(r, num_iterations, base_iter, base_limit);
#endif

	base_iter += num_iterations;
	while (base_limit < base_iter)
		base_limit *= 2;

	r->curr_iter = base_iter;
	r->curr_limit = base_limit;
}

/*---------------------------------------------------------------------*/
static void
packed_to_mp(uint32 *x, uint32 stride, mp_t *res)
{
	uint32 i, j;

	for (i = j = 0; i < MAX_MP_WORDS; i++, j+=stride)
		res->val[i] = x[j];

	for (i = MAX_MP_WORDS; i; i--) {
		if (res->val[i-1])
			break;
	}
	res->nwords = i;
}

/*---------------------------------------------------------------------*/
mp_t *
rho_get_curr(rho_t *r)
{
	uint32 i, j, k;
	uint32 n = r->num_functions;

#ifdef HAVE_CUDA
	/* pull the accumulator off the card */

	CUDA_TRY(cuMemcpyDtoH(r->main_array, r->gpu_iter_data, 
				2 * MAX_MP_WORDS * n * sizeof(uint32)))

	/* manually convert out of Montgomery form */

	for (i = 0; i < n; i++) {

		r->gcd_array[i] = 1;
		for (j = 1, k = n; j < MAX_MP_WORDS; j++, k += n)
			r->gcd_array[k+i] = 0;

		montmul(r->accum_array + i, r->gcd_array + i,
			r->mont_temp + i, n,
			r->target.val, r->mont_w);

		for (j = k = 0; j < MAX_MP_WORDS; j++, k += n)
			r->gcd_array[k+i] = r->mont_temp[k+i];
	}

	/* copy the converted accumulator values back onto the card */

	CUDA_TRY(cuMemcpyHtoD(r->gpu_iter_data, 
			r->main_array, MAX_MP_WORDS * 
			n * sizeof(uint32)))

	/* compute the GCD */

	CUDA_TRY(cuLaunchGrid(r->gpu_kernel2, r->num_blocks, 1))

	/* retrieve the result */

	CUDA_TRY(cuMemcpyDtoH(r->main_array, r->gpu_iter_data, 
				MAX_MP_WORDS * n * sizeof(uint32)))
#else

	for (i = 0; i < n; i++) {

		r->gcd_array[i] = 1;
		for (j = 1, k = n; j < MAX_MP_WORDS; j++, k += n)
			r->gcd_array[k+i] = 0;

		montmul(r->accum_array + i, r->gcd_array + i,
			r->mont_temp + i, n,
			r->target.val, r->mont_w);

		for (j = k = 0; j < MAX_MP_WORDS; j++, k += n) {
			r->gcd_array[k+i] = r->mont_temp[k+i];
			r->mont_temp[k+i] = r->target.val[j];
		}

		gcd(r->mont_temp + i, r->gcd_array + i, n);
	}

#endif

	for (i = 0; i < n; i++) {
		packed_to_mp(r->gcd_array + i, n, r->curr_values + i);
	}

	return r->curr_values;
}
