/*----------------------------------------------------------------------
 * This source distribution and its contents are (c) 2011 Gregory Perry, 
 * and are released under the GNU Public License Version 3.  Portions of
 * this distribution were borrowed from the msieve project.
 *
 * For more information visit the main CloudCrack page at:
 * http://www.GoVirtual.tv/CloudCrack/
 *
 * Become of fan and engage in discussions at the CloudCrack Fan Page:
 * https://www.facebook.com/pages/CloudCrack/156984984362011/
 *
 * Or, send an email to Gregory.Perry@GoVirtual.tv
 *
 * $Id$
 * --------------------------------------------------------------------*/

#include "cudarho_priv.h"

/*---------------------------------------------------------------------*/
int32
mont_cmp(uint32 *a, uint32 stride_a,
	 uint32 *b, uint32 stride_b)
{
	uint32 i, j;

	i = stride_a * (MAX_MP_WORDS - 1);
	j = stride_b * (MAX_MP_WORDS - 1);
	while ((int32)j >= 0) {
		uint32 ai = a[i];
		uint32 bj = b[j];

		if (ai > bj)
			return 1;
		else if (ai < bj)
			return -1;

		i -= stride_a;
		j -= stride_b;
	}

	return 0;
}

/*---------------------------------------------------------------------*/
uint32
mont_add(uint32 *a, uint32 stride_a,
		uint32 *b, uint32 stride_b,
		uint32 *c, uint32 stride_c)
{
	uint32 i, j, k, m;
	int64 carry = 0;

	for (i = j = k = m = 0; m < MAX_MP_WORDS; m++) {

		uint64 ai = a[i];
		uint64 bj = b[j];

		carry += ai + bj;
		c[k] = (uint32)carry;
		carry = carry >> 32;

		i += stride_a;
		j += stride_b;
		k += stride_c;
	}

	return carry;
}

/*---------------------------------------------------------------------*/
void
mont_sub(uint32 *a, uint32 stride_a,
		uint32 *b, uint32 stride_b,
		uint32 *c, uint32 stride_c)
{
	uint32 i, j, k, m;
	int64 carry = 0;

	for (i = j = k = m = 0; m < MAX_MP_WORDS; m++) {

		uint64 ai = a[i];
		uint64 bj = b[j];

		carry += ai - bj;
		c[k] = (uint32)carry;
		carry = carry >> 32;

		i += stride_a;
		j += stride_b;
		k += stride_c;
	}
}

/*---------------------------------------------------------------------*/
void 
montmul(uint32 *a, uint32 *b, uint32 *scratch,
		uint32 stride, uint32 *n, uint32 w)
{
	uint32 i, j, k, m;

	for (i = k = 0; i <= MAX_MP_WORDS; i++, k+=stride)
		scratch[k] = 0;

	for (i = k = 0; i < MAX_MP_WORDS; i++, k+=stride) {

		uint32 q;
		uint32 carry;
		uint64 accum;

		q = a[k];
		accum = (uint64)q * b[0] + scratch[0];
		scratch[0] = (uint32)accum;
		carry = (uint32)(accum >> 32);

		for (j = 1, m = stride; j < MAX_MP_WORDS; j++, m+=stride) {
			accum = (uint64)q * b[m] + carry + scratch[m];
			scratch[m] = (uint32)accum;
			carry = (uint32)(accum >> 32);
		}

		scratch[m] += carry;
		q = scratch[0] * w;
		accum = (uint64)q * n[0] + scratch[0];
		carry = (uint32)(accum >> 32);

		for (j = 1, m = stride; j < MAX_MP_WORDS; j++, m+=stride) {
			accum = (uint64)q * n[j] + carry + scratch[m];
			scratch[m - stride] = (uint32)accum;
			carry = (uint32)(accum >> 32);
		}

		accum = (uint64)scratch[m] + carry;
		scratch[m - stride] = (uint32)accum;
		scratch[m] = (uint32)(accum >> 32);
	}

	if (scratch[m] || mont_cmp(scratch, stride, n, 1) >= 0)
		mont_sub(scratch, stride, n, 1, scratch, stride);
}
/*---------------------------------------------------------------------*/
static uint32
mont_is_zero(uint32 *a, uint32 stride_a)
{
	uint32 i, j;

	for (i = j = 0; i < MAX_MP_WORDS; i++, j += stride_a) {
		if (a[j] != 0)
			return 0;
	}

	return 1;
}

/*---------------------------------------------------------------------*/
void
gcd(uint32 *a, uint32 *b, uint32 stride)
{
	uint32 i, j;
	uint32 *u = a;
	uint32 *v = b;

	if (mont_is_zero(v, stride)) {
		return;
	}
	else if (v[0] & 1) {
		u = b;
		v = a;
		mont_sub(v, stride, u, stride, v, stride);
	}

	do {
		do {
			uint32 shift, comp_shift;
			uint32 lo, hi;

			shift = ffs(v[0]) - 1;
			if (shift > 31)
				shift = 31;

			comp_shift = 32 - shift;
			lo = v[0];
			for (i = j = 0; i < MAX_MP_WORDS-1; i++, j+=stride) {
				hi = v[j+stride];
				v[j] = lo >> shift | hi << comp_shift;
				lo = hi;
			}
			v[j] = lo >> shift;

		} while ((v[0] & 1) == 0 && !mont_is_zero(v, stride));

		if (mont_cmp(u, stride, v, stride) > 0) {
			uint32 *tmp = u;
			u = v;
			v = tmp;
		}

		mont_sub(v, stride, u, stride, v, stride);

	} while (!mont_is_zero(v, stride));

	if (mont_is_zero(b, stride)) {
		for (i = j = 0; i < MAX_MP_WORDS; i++, j+=stride)
			b[j] = a[j];
	}
}

