/*----------------------------------------------------------------------
 * This source distribution and its contents are (c) 2011 Gregory Perry, 
 * and are released under the GNU Public License Version 3.  Portions of
 * this distribution were borrowed from the msieve project.
 *
 * For more information visit the main CloudCrack page at:
 * http://www.GoVirtual.tv/CloudCrack/
 *
 * Become of fan and engage in discussions at the CloudCrack Fan Page:
 * https://www.facebook.com/pages/CloudCrack/156984984362011/
 *
 * Or, send an email to Gregory.Perry@GoVirtual.tv
 *
 * $Id$
 * --------------------------------------------------------------------*/

#include <cudarho.h>

/*--------------------------------------------------------------------*/
void get_random_seeds(uint32 *seed1, uint32 *seed2) {

	uint32 tmp_seed1, tmp_seed2;

	/* In a multithreaded program, every msieve object
	   should have two unique, non-correlated seeds
	   chosen for it */

#if !defined(WIN32) && !defined(_WIN64)

	FILE *rand_device = fopen("/dev/urandom", "r");

	if (rand_device != NULL) {

		/* Yay! Cryptographic-quality nondeterministic randomness! */

		fread(&tmp_seed1, sizeof(uint32), (size_t)1, rand_device);
		fread(&tmp_seed2, sizeof(uint32), (size_t)1, rand_device);
		fclose(rand_device);
	}
	else

#endif
	{
		/* <Shrug> For everyone else, sample the current time,
		   the high-res timer (hopefully not correlated to the
		   current time), and the process ID. Multithreaded
		   applications should fold in the thread ID too */

		uint64 high_res_time = read_clock();
		tmp_seed1 = ((uint32)(high_res_time >> 32) ^
			     (uint32)time(NULL)) * 
			    (uint32)getpid();
		tmp_seed2 = (uint32)high_res_time;
	}

	/* The final seeds are the result of a multiplicative
	   hash of the initial seeds */

	(*seed1) = tmp_seed1 * ((uint32)40499 * 65543);
	(*seed2) = tmp_seed2 * ((uint32)40499 * 65543);
}

/*--------------------------------------------------------------------*/
void print_usage(char *progname) {

	printf("\nusage: %s [options] [one_number]\n", progname);
	printf("\nnumbers starting with '0' are treated as octal,\n"
		"numbers starting with '0x' are treated as hexadecimal\n");
	printf("\noptions:\n"
	         "   -i <num> number of rho iterations\n"
	         "   -f <num> number of rho functions\n"
	         "   -b <num> iterations between GCDs\n"
	         "   -l <num> number of bits in initial values\n"
	         "   -m <num> residue modulo 9 of initial values\n"
#ifdef HAVE_CUDA
		 "   -g <num>  use GPU <num>, 0 <= num < (# graphics cards)>\n"
#endif
	      );
}

/*--------------------------------------------------------------------*/
typedef struct {
	uint32 seed1;
	uint32 seed2;
	uint32 num_bits;
	uint32 resclass;
} rng_state_t;

static INLINE uint32 
myrand(uint32 *rand_seed, uint32 *rand_carry) 
{
	/* A multiply-with-carry generator by George Marsaglia.
	   The period is about 2^63. */

	#define RAND_MULT 2131995753

	uint64 temp = (uint64)(*rand_seed) * 
		       (uint64)RAND_MULT + 
		       (uint64)(*rand_carry);
	*rand_seed = (uint32)temp;
	*rand_carry = (uint32)(temp >> 32);
	return (uint32)temp;
}

void myrng(uint32 which_rho, void *s, mp_t *out)
{
	rng_state_t *ss = (rng_state_t *)s;

	mp_rand(ss->num_bits, out, &ss->seed1, &ss->seed2);
	mp_mul_1(out, 9, out);
	mp_add_1(out, ss->resclass, out);
}

/*--------------------------------------------------------------------*/
int main(int argc, char **argv) {

	char buf[400];
	rng_state_t state;
	uint32 which_gpu = 0;
	uint32 num_iter = 1000;
	uint32 num_functions = 5;
	uint32 batch_size = 100;
	uint32 num_iter_done = 0;
	int32 i;
	rho_t rho;
	mp_t *outvals;

	state.num_bits = 32;
	state.resclass = 5;
		
	i = 1;
	buf[0] = 0;
	while (i < argc) {
		if (argv[i][0] == (char)('-')) {
			switch(tolower(argv[i][1])) {
			case 'h':
			case '?':
				print_usage(argv[0]);
				return 0;

#ifdef HAVE_CUDA					
			case 'g':
				if (i + 1 < argc && isdigit(argv[i+1][0])) {
					which_gpu = atol(argv[i+1]);
					i += 2;
				}
				else {
					print_usage(argv[0]);
					return -1;
				}
				break;
#endif					
			case 'i':
				if (i + 1 < argc && isdigit(argv[i+1][0])) {
					num_iter = atol(argv[i+1]);
					i += 2;
				}
				else {
					print_usage(argv[0]);
					return -1;
				}
				break;

			case 'b':
				if (i + 1 < argc && isdigit(argv[i+1][0])) {
					batch_size = atol(argv[i+1]);
					i += 2;
				}
				else {
					print_usage(argv[0]);
					return -1;
				}
				break;

			case 'l':
				if (i + 1 < argc && isdigit(argv[i+1][0])) {
					state.num_bits = atol(argv[i+1]);
					i += 2;
				}
				else {
					print_usage(argv[0]);
					return -1;
				}
				break;

			case 'm':
				if (i + 1 < argc && isdigit(argv[i+1][0])) {
					state.resclass = atol(argv[i+1]);
					i += 2;
				}
				else {
					print_usage(argv[0]);
					return -1;
				}
				break;

			case 'f':
				if (i + 1 < argc && isdigit(argv[i+1][0])) {
					num_functions = atol(argv[i+1]);
					i += 2;
				}
				else {
					print_usage(argv[0]);
					return -1;
				}
				break;

			default:
				print_usage(argv[0]);
				return -1;
			}
		}
		else {
			if (isdigit(argv[i][0]) || argv[i][0] == '(' )
				strncpy(buf, argv[i], sizeof(buf));
			i++;
		}
	}

	if (buf[0] == 0) {
		print_usage(argv[0]);
		return -1;
	}

	get_random_seeds(&state.seed1, &state.seed2);

	rho_init(&rho, buf, myrng, &state, num_functions, which_gpu);

	num_functions = rho.num_functions;

	while (num_iter_done < num_iter) {

		uint32 curr_iter = MIN(batch_size, 
				num_iter - num_iter_done);

		rho_run(&rho, curr_iter);

		num_iter_done += curr_iter;
		printf("%u iterations\n", num_iter_done);

		outvals = rho_get_curr(&rho);

		for (i = 0; i < (int32)num_functions; i++) {
			if (!mp_is_one(outvals + i) &&
			    mp_cmp(&rho.target, outvals + i)) {
				printf("function %u found factor %s\n", i,
					mp_sprintf(outvals + i, 10, buf));
				goto cleanup;
			}
		}
	}

cleanup:
	rho_free(&rho);

	return 0;
}
